(* $Id: protocol.ml 953 2011-09-02 21:28:08Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_exec

let lf_re = Str.regexp "\n";;

let tab_re = Str.regexp "[\t]";;

class display_protocol ?(auto_update=false) (pg : Miniui.page) =
  let concmd_buf = Netbuffer.create 50 in
object(self)

  val mutable buffer = Buffer.create 1000 
			 
  val width = pg # width()

  val mutable current_line = None
	      (* None: no current line
	       * Some s: The current line is already printed 
	       *)
  val mutable logical_start = true
	      (* The next line will be a the beginning of a new
	       * logical line
	       *)

  method add_string s =
    Netbuffer.add_string concmd_buf s;
    match Godi_concmd.scan concmd_buf false with
      | `Data d ->
	  self # add_data d
      | `Incomplete ->
	  ()
      | `Prompt(p, f) ->
	  self # prompt false p f
      | `Password_prompt(p, f) ->
	  self # prompt true p f


  method private prompt is_password p f =
    pg # put_line "### <<< User interaction required >>> ###";
    pg # activate_text_entry 
      ~hidden:is_password
      ~prompt:p
      ~callback_ok:(fun s ->
		      Godi_concmd.send_data f s;
		      if not is_password then
			self # add_data(p ^ s ^ "\n")
		   )
      ~callback_cancel:(fun () ->
			  (* these is no cancel, so restart *)
			  self # prompt is_password p f
		       )
      ()


  method private add_data s =
    Buffer.add_string buffer s;
    let s = Str.global_replace tab_re " " s in  (* replace TABs with SPACEs *)
    let segs = Str.split_delim lf_re s in
    let n_segs = List.length segs in
    let n = ref 0 in
    List.iter
      (fun seg ->
	 let log_line_ends = !n < n_segs - 1 in
	     (* [seg] is the last segment of the current logical line *)
	 let seg_rest = ref seg in
	 while !seg_rest <> "" do
	   let line, is_appended =
	     match current_line with
		 None -> 
		   ((if logical_start then "> " else "  "), true)
	       | Some l ->
		   (l, false)
	   in
	   let w = width - String.length line in
	   assert(w > 0);
	   let piece, new_rest =
	     if String.length !seg_rest < w then
	       (!seg_rest, "")
	     else
	       (String.sub !seg_rest 0 w,
		String.sub !seg_rest w (String.length !seg_rest - w))
	   in
	   let line' = line ^ piece in

	   (* Output line': *)
	   if is_appended then
	     pg # put_line ~line:`Append line'
	   else
	     pg # put_line ~line:(`Replace(pg # length() - 1)) line';

	   (* If the line is full, begin the next line, else remember
	    * the current line for continuation
	    *)
	   if String.length line' >= width then
	     current_line <- None
	   else
	     current_line <- Some line';

	   seg_rest := new_rest;
	   logical_start <- false;

	 done;
	 if log_line_ends then (
	   current_line <- None;
	   logical_start <- true
	 );
	 incr n;
      )
      segs;
    if auto_update then pg#update();

  method collector : collector =
    `Forward (self#add_string)

  method buffer = buffer


  method save_dialog() =
    let cwd = Sys.getcwd() in
    pg # activate_text_entry 
      ~value:cwd
      ~prompt:"Save as: "
      ~callback_ok:(fun s ->
		      try
			self # write_to_file s
		      with
			  error ->
			    pg # configure ~auto_scroll:true ();
			    let msg = Printexc.to_string error in
			    pg # put_line "";
			    pg # put_line ("ERROR: " ^ msg))
      ~callback_cancel:(fun () -> ())
      ()
    
  method private write_to_file filename =
    let f = Godi_file.open_out `Binary (Godi_file.filename filename) in
    try
      Buffer.output_buffer f buffer;
      close_out f
    with
	error -> close_out f; raise error

end
;;
